import contextlib
import lzma
import os
import re
import shutil
import tempfile
import xml.etree.ElementTree as ElementTree

import click
import delegator
import requests
import semver

from .base import BasePlatformGadget, BasePlatformPatcher, objection_path
from .github import Github
from ..helpers import debug_print


class AndroidGadget(BasePlatformGadget):
    """ Class used to download Android Frida libraries """

    android_library_path = os.path.join(objection_path, 'android')

    # Lists the supported architectures. Key matches Android support
    # https://developer.android.com/ndk/guides/abis.html#sa
    # Value matches library arch for frida.
    architectures = {
        'armeabi': 'arm',
        'armeabi-v7a': 'arm',
        'arm64': 'arm64',
        'arm64-v8a': 'arm64',
        'x86': 'x86',
        'x86_64': 'x86_64',
    }

    def __init__(self, github: Github) -> None:
        """
            Build a new instance, ensuring that the paths needed
            are available.

            :param github:
        """

        super(AndroidGadget, self).__init__(github)

        self.architecture = None

        # prep paths. if they dont exist, create them
        for path in self.architectures.keys():

            d = os.path.join(self.android_library_path, path)

            if not os.path.exists(d):
                os.makedirs(d)

    def set_architecture(self, architecture: str):
        """
            Set the CPU architecture we will work with.

            :param architecture:
            :return:
        """

        if architecture not in self.architectures.keys():
            raise Exception('Invalid architecture `{0}` set. Valid options are: {1}'.format(
                architecture, ', '.join(self.architectures)))

        self.architecture = architecture

        return self

    def get_architecture(self) -> str:
        """
            Get the architecture we are working with.

            :return:
        """

        return self.architecture

    def get_frida_library_path(self, packed: bool = False) -> str:
        """
            Get the path to a frida-library, both in the packed and

            :param packed:
            :return:
        """

        if not self.architecture:
            raise Exception('Unable to determine path without architecture')

        return os.path.join(self.android_library_path, self.architecture,
                            'libfrida-gadget.so' + ('.xz' if packed else ''))

    def gadget_exists(self) -> bool:
        """
            Determines of a frida-gadget library exists.

            :return:
        """

        if not self.architecture:
            raise Exception('Unable to determine path without architecture')

        return os.path.exists(self.get_frida_library_path())

    def download(self):
        """
            Downloads the latest Android gadget for this
            architecture.

            :return:
        """

        download_url = self._get_download_url()
        click.secho('Downloading from: {0}'.format(download_url), dim=True)

        # stream the download using requests
        library = requests.get(download_url, stream=True)
        library_destination = self.get_frida_library_path(packed=True)

        # save the requests stream to file
        with open(library_destination, 'wb') as f:
            click.secho('Downloading {0} library to {1}...'.format(self.architecture,
                                                                   library_destination), fg='green', dim=True)

            shutil.copyfileobj(library.raw, f)

        return self

    def _get_download_url(self) -> str:
        """
            Determines the download URL to use for the Android
            gadget.

            :return:
        """

        url = ''

        # url should contain 'frida-gadget-{version}-android-{arch}.so.xz
        url_start = 'frida-gadget-'
        url_end = 'android-' + self.architectures[self.architecture] + '.so.xz'

        for asset in self.github.get_assets():
            if asset['name'].startswith(url_start) and asset['name'].endswith(url_end):
                url = asset['browser_download_url']

        if not url:
            click.secho('Unable to determine URL to download the library', fg='red')
            raise Exception('Unable to determine URL for Android gadget download.')

        return url

    def unpack(self):
        """
            Unpacks a downloaded .xz gadget.

            :return:
        """

        click.secho('Unpacking {0}...'.format(self.get_frida_library_path(packed=True)), dim=True)

        with lzma.open(self.get_frida_library_path(packed=True)) as f:
            with open(self.get_frida_library_path(), 'wb') as g:
                g.write(f.read())

        return self

    def cleanup(self):
        """
            Cleans up a downloaded iOS .xz gadget.

            :return:
        """

        click.secho('Cleaning up downloaded archives...', dim=True)

        os.remove(self.get_frida_library_path(packed=True))


class AndroidPatcher(BasePlatformPatcher):
    """ Class used to patch Android APK's"""

    required_commands = {
        'aapt': {
            'installation': 'apt install aapt (Kali Linux)'
        },
        'adb': {
            'installation': 'apt install adb (Kali Linux); brew install adb (macOS)'
        },
        'apksigner': {
            'installation': 'apt install apksigner (Kali Linux)'
        },
        'apktool': {
            'installation': 'Install from https://apktool.org/docs/install'
        },
        'zipalign': {
            'installation': 'apt install zipalign (Kali Linux)'
        }
    }

    def __init__(self, skip_cleanup: bool = False, skip_resources: bool = False, manifest: str = None, only_main_classes: bool = False):
        super(AndroidPatcher, self).__init__()

        self.apk_source = None
        self.apk_temp_directory = tempfile.mkdtemp(suffix='.apktemp')
        self.apk_temp_frida_patched = self.apk_temp_directory + '.objection.apk'
        self.apk_temp_frida_patched_aligned = self.apk_temp_directory + '.aligned.objection.apk'
        self.aapt = None
        self.skip_cleanup = skip_cleanup
        self.skip_resources = skip_resources
        self.manifest = manifest

        self.keystore = os.path.join(os.path.abspath(os.path.dirname(__file__)), '../assets', 'objection.jks')
        self.netsec_config = os.path.join(os.path.abspath(os.path.dirname(__file__)), '../assets',
                                          'network_security_config.xml')
        self.only_main_classes = only_main_classes

    def is_apktool_ready(self) -> bool:
        """
            Check if apktool is ready for use.

            :return:bool
        """

        min_version = '2.6.0'  # the version of apktool we require

        o = delegator.run(self.list2cmdline([
            self.required_commands['apktool']['location'],
            '-version',
        ]), timeout=self.command_run_timeout).out.strip()

        # On windows we get this 'Press any key to continue' thing,
        # localized to the the current language. Assume that the version
        # string we want is always the first line.
        if len(o.split('\n')) > 1:
            o = o.split('\n')[0]

        # Apktool v2.12.0 has changed the syntax `apktool -version, this grabs the version from the usage screen output
        # instead of re-running as `apktool v`.
        if len(o.split(' ')) > 1:
            o = o.split(' ')[1]

        if len(o) == 0:
            click.secho('Unable to determine apktool version. Is it installed')
            return False

        click.secho('Detected apktool version as: ' + o, dim=True)

        # ensure we have at least apktool MIN_VERSION
        if semver.compare(o, min_version) < 0:
            click.secho('apktool version should be at least ' + min_version, fg='red', bold=True)
            click.secho('Please see the following URL for more information: '
                        'https://github.com/sensepost/objection/wiki/Apktool-Upgrades', fg='yellow')
            return False

        # run clean-frameworks-dir
        click.secho('Running apktool empty-framework-dir...', dim=True)
        o = delegator.run(self.list2cmdline([
            self.required_commands['apktool']['location'],
            'empty-framework-dir',
        ]), timeout=self.command_run_timeout).out.strip()

        if len(o) > 0:
            click.secho(o, fg='yellow', dim=True)

        return True

    def set_apk_source(self, source: str):
        """
            Set the source APK to work with.

            :param source:
            :return:
        """

        if not os.path.exists(source):
            raise Exception('Source {0} not found.'.format(source))

        self.apk_source = source

        return self

    def _get_android_manifest(self) -> ElementTree:
        """
            Get the AndroidManifest as a parsed ElementTree

            :return:
        """

        # error if --skip-resources was used because the manifest is encoded
        if self.skip_resources is True and self.manifest is None:
            click.secho('Cannot manually parse the AndroidManifest.xml when --skip-resources '
                        'is set, remove this and try again, or manually specify a manifest with --manifest.', fg='red')
            raise Exception('Cannot --skip-resources when trying to manually parse the AndroidManifest.xml')

        # use the android namespace
        ElementTree.register_namespace('android', 'http://schemas.android.com/apk/res/android')
        if self.manifest is not None:
            return ElementTree.parse(self.manifest)
        else:
            return ElementTree.parse(os.path.join(self.apk_temp_directory, 'AndroidManifest.xml'))

    def _get_appt_output(self):
        """
            Get the output of `aapt dump badging`.

            :return:
        """

        if not self.aapt:
            o = delegator.run(self.list2cmdline([
                self.required_commands['aapt']['location'],
                'dump',
                'badging',
                self.apk_source
            ]), timeout=self.command_run_timeout)

            if len(o.err) > 0:
                click.secho('An error may have occurred while running aapt.', fg='red')
                click.secho(o.err, fg='red')

            self.aapt = o.out

        return self.aapt

    def _get_launchable_activity(self) -> str:
        """
            Determines the class name for the activity that is
            launched on application startup.

            This is done by first trying to parse the output of
            aapt dump badging, then falling back to manually
            parsing the AndroidManifest for activity-alias tags.

            :return:
        """

        activities = (match.groups()[0] for match in
                      re.finditer(r"^launchable-activity: name='([^']+)'", self._get_appt_output(), re.MULTILINE))
        activity = next(activities, None)

        # If we got the activity using aapt, great, return that
        if activity is not None:
            return activity

        # if we dont have the activity yet, check out activity aliases
        click.secho(('Unable to determine the launchable activity using aapt, trying '
                     'to manually parse the AndroidManifest for activity aliases...'), dim=True, fg='yellow')

        # Try and parse the manifest manually
        manifest = self._get_android_manifest()
        root = manifest.getroot()

        # grab all of the activity-alias tags
        for alias in root.findall('./application/activity-alias'):

            # Take not of the current activity
            current_activity = alias.get('{http://schemas.android.com/apk/res/android}targetActivity')
            categories = alias.findall('./intent-filter/category')

            # make sure we have categories for this alias
            if categories is None:
                continue

            for category in categories:

                # check if the name of this category is that of LAUNCHER
                # its possible to have multiples, but once we determine one
                # that fits we can just return and move on
                category_name = category.get('{http://schemas.android.com/apk/res/android}name')

                if category_name == 'android.intent.category.LAUNCHER':
                    return current_activity

        # getting here means we were unable to determine what the launchable
        # activity is
        click.secho('Unable to determine the launchable activity for this app.', fg='red')
        raise Exception('Unable to determine launchable activity')

    def get_patched_apk_path(self) -> str:
        """
            Returns the path of the patched, aligned APK.

            :return:
        """

        return self.apk_temp_frida_patched_aligned

    def get_temp_working_directory(self) -> str:
        """
            Returns the temporary working directory used by this patcher.

            :return:
        """

        return self.apk_temp_directory

    def unpack_apk(self, fix_concurrency_to = None):
        """
            Unpack an APK with apktool.

            :return:
        """

        click.secho('Unpacking {0}'.format(self.apk_source), dim=True)


        o = delegator.run(self.list2cmdline([
            self.required_commands['apktool']['location'],
            'decode',
            '-f',
        ] +
          (['-r'] if self.skip_resources else []) +
          (['--only-main-classes'] if self.only_main_classes else []) +
        [
            '-o',
            self.apk_temp_directory,
            self.apk_source
        ] + ([] if fix_concurrency_to is None else ['-j', fix_concurrency_to])), timeout=self.command_run_timeout)

        debug_print("Command:" + o.cmd)

        if len(o.err) > 0:
            click.secho('An error may have occurred while extracting the APK.', fg='red')
            click.secho(o.err, fg='red')
            
    def inject_internet_permission(self):
        """
            Checks the status of the source APK to see if it
            has the INTERNET permission. If not, the manifest file
            is parsed and the permission injected.

            :return:
        """

        internet_permission = 'android.permission.INTERNET'

        # if the app already has the internet permission, easy mode :D
        if internet_permission in self._get_appt_output():
            click.secho('App already has android.permission.INTERNET', fg='green')
            return

        # if not, we need to inject an element with it
        click.secho('App does not have android.permission.INTERNET, attempting to patch the AndroidManifest.xml...',
                    dim=True, fg='yellow')
        xml = self._get_android_manifest()
        root = xml.getroot()

        click.secho('Injecting permission: {0}'.format(internet_permission), fg='green')

        # prepare a new 'uses-permission' tag
        child = ElementTree.Element('uses-permission')
        child.set('android:name', internet_permission)
        root.append(child)

        click.secho('Writing new Android manifest...', dim=True)

        xml.write(os.path.join(self.apk_temp_directory, 'AndroidManifest.xml'),
                  encoding='utf-8', xml_declaration=True)

    def extract_native_libs_patch(self):
        """
            Check the AndroidManifest.xml file for extractNativeLibs="false"
            if it exists, change it to extractNativeLibs="true".

            Since AndroidStudio 2.1 this flag is set as false by default.
            This breaks it when installing the .apk to the device.

            :return:
        """
        xml = self._get_android_manifest()
        root = xml.getroot()

        application_tag = root.findall('application')

        # ensure that we got the application tag
        if len(application_tag) <= 0:
            message = 'Could not find the application tag in the AndroidManifest.xml'
            click.secho(message, fg='red', bold=True)
            raise Exception(message)

        application_tag = application_tag[0]

        # Check if the flag is present and set to false
        if '{http://schemas.android.com/apk/res/android}extractNativeLibs' in application_tag.attrib \
                and application_tag.attrib['{http://schemas.android.com/apk/res/android}extractNativeLibs'] == 'false':
            # Set the flag to true
            application_tag.attrib['{http://schemas.android.com/apk/res/android}extractNativeLibs'] = 'true'
            click.secho('Setting extractNativeLibs to true...', dim=True)
            xml.write(os.path.join(self.apk_temp_directory, 'AndroidManifest.xml'),
                      encoding='utf-8', xml_declaration=True)
            return

    def flip_debug_flag_to_true(self):
        """
            Set the android:debuggable flag to true in the
            AndroidManifest.

            :return:
        """

        xml = self._get_android_manifest()
        root = xml.getroot()

        click.secho('Setting debug flag to true', fg='green')

        application_tag = root.findall('application')

        # ensure that we got the application tag
        if len(application_tag) <= 0:
            message = 'Could not find the application tag in the AndroidManifest.xml'
            click.secho(message, fg='red', bold=True)
            raise Exception(message)

        application_tag = application_tag[0]

        if '{http://schemas.android.com/apk/res/android}debuggable' in application_tag.attrib \
                and application_tag.attrib['{http://schemas.android.com/apk/res/android}debuggable'] == 'true':
            click.secho('Application already has the android:debuggable flag set to True')
            return

        # set the debuggable flag
        application_tag.attrib['{http://schemas.android.com/apk/res/android}debuggable'] = 'true'

        click.secho('Writing new Android manifest...', dim=True)
        xml.write(os.path.join(self.apk_temp_directory, 'AndroidManifest.xml'),
                  encoding='utf-8', xml_declaration=True)

    def add_network_security_config(self):
        """
            Add a network_security_config.xml to the AndroidManifest.xml for
            Android 7+.

            Refs:
                https://serializethoughts.com/2016/09/10/905/
                https://warroom.securestate.com/android-7-intercepting-app-traffic/
                https://www.nowsecure.com/blog/2017/06/15/certificate-pinning-for-android-and-ios-mobile-man-in-the-middle-attack-prevention/
                https://android-developers.googleblog.com/2016/07/changes-to-trusted-certificate.html
                https://www.nccgroup.trust/uk/about-us/newsroom-and-events/blogs/2017/november/bypassing-androids-network-security-configuration/
                https://sensepost.com/blog/2018/tip-toeing-past-android-7s-network-security-configuration/

            return:
        """

        xml = self._get_android_manifest()
        root = xml.getroot()
        application_tag = root.findall('application')

        # ensure that we got the application tag
        if len(application_tag) <= 0:
            message = 'Could not find the application tag in the AndroidManifest.xml'
            click.secho(message, fg='red', bold=True)
            raise Exception(message)

        application_tag = application_tag[0]

        click.secho('Checking for an existing networkSecurityConfig tag', dim=True)

        if '{http://schemas.android.com/apk/res/android}networkSecurityConfig' in application_tag.attrib:
            if not click.prompt('An existing network security config was found. Do you want to replace it?',
                                type=bool, default=True):
                return

        # copy our network security configuration to res/xml/network_security_config.xml
        sec_config_path = os.path.join(self.apk_temp_directory, 'res', 'xml')

        # check if the config path exists
        if not os.path.exists(sec_config_path):
            click.secho('Creating XML res path: {0}'.format(sec_config_path), dim=True)
            os.makedirs(sec_config_path)

        click.secho('Copying network_security_config.xml...', fg='green', dim=True)
        shutil.copyfile(self.netsec_config, os.path.join(sec_config_path, 'network_security_config.xml'))

        # set the networkSecurityConfig xml location
        # this is in res/xml/network_security_config.xml
        application_tag.attrib[
            '{http://schemas.android.com/apk/res/android}networkSecurityConfig'] = '@xml/network_security_config'

        click.secho('Writing new Android manifest...', dim=True)
        xml.write(os.path.join(self.apk_temp_directory, 'AndroidManifest.xml'),
                  encoding='utf-8', xml_declaration=True)

    def _determine_smali_path_for_class(self, target_class) -> str:
        """
            Attempts to determine the local path for a target class' smali

            :param target_class:
            :return:
        """

        # convert to a filesystem path, just like how it would be on disk
        # from the apktool dump
        target_class = target_class.replace('.', '/')

        activity_path = os.path.join(self.apk_temp_directory, 'smali', target_class) + '.smali'

        # check if the activity path exists. If not, try and see if this may have been
        # a multidex setup
        if not os.path.exists(activity_path):

            click.secho('Smali not found in smali directory. This might be a multidex APK. Searching...', dim=True)

            # apk tool will dump the dex classes to a smali directory. in multidex setups
            # we have folders such as smali_classes2, smali_classes3 etc. we will search
            # those paths for the launch activity we detected.
            for x in range(2, 100):
                smali_path = os.path.join(self.apk_temp_directory, 'smali_classes{0}'.format(x))

                # stop if the smali_classes directory does not exist.
                if not os.path.exists(smali_path):
                    break

                # determine the path to the launchable activity again
                activity_path = os.path.join(smali_path, target_class) + '.smali'

                # if we found the activity, stop the loop
                if os.path.exists(activity_path):
                    click.secho('Found smali at: {0}'.format(activity_path), dim=True)
                    break

        # one final check to ensure we have the target .smali file
        if not os.path.exists(activity_path):
            raise Exception('Unable to find smali to patch!')

        return activity_path

    @staticmethod
    def _determine_end_of_smali_method_from_line(smali: list, start: int) -> int:
        """
            Determines where the .end method line is.

            This method is also aware of a methods that 'returns' and will
            return the line before that too.

            :param smali:
            :param start:
            :return:
        """

        # enumerate all of # the lines in the original smali sources and mark the offsets of the
        # lines that contain '.end method'. the search starts right after the
        # original inject marker so that we can pick the top most .end method
        # when we are done searching. this is also why the # represented in the
        # inject marker is added to the calculated marker in the list of end methods.
        end_methods = [(i + start) for i, x in enumerate(smali[start:]) if '.end method' in x]

        # ensure that we found at least one .end method
        if len(end_methods) <= 0:
            raise Exception('Unable to find the end of the existing constructor')

        # set the last line of the constructors method to the one
        # just before the .end method line
        end_of_method = end_methods[0] - 1

        # check if the constructor has a return type call. if it does,
        # move up one line again to inject our loadLibrary before the return
        if 'return' in smali[end_of_method]:
            end_of_method -= 1

        return end_of_method

    @staticmethod
    def _determine_first_inject_point_of_smali_method_from_line(smali: list, start: int) -> int:
        """
            Determines the first line in a smali method where we can inject code.
            This is the line after any .locals or .annotations

            This method is also aware of a methods that 'returns' and will
            return the line before that too.

            :param smali:
            :param start:
            :return:
        """

        pos = start
        in_annotation = False
        while pos + 1 < len(smali):
            pos = pos + 1
            line = smali[pos].strip()

            # skip empty lines
            if not line:
                continue

            # skip locals
            if line.startswith(".locals "):
                continue

            # skip annotations
            if in_annotation or line.startswith(".annotation "):
                in_annotation = True
                continue

            if line.startswith(".end annotation"):
                in_annotation = False
                continue

            return pos - 1

    def _patch_smali_with_load_library(self, smali_lines: list, inject_marker: int) -> list:
        """
            Patches a list of smali lines with the appropriate
            loadLibrary call based on wether a constructor already
            exists or not.

            :param smali_lines:
            :param inject_marker:
            :return:
        """

        # raw smali to inject.
        # ref: https://koz.io/using-frida-on-android-without-root/

        # if no constructor is present, the full_load_library is used
        full_load_library = ('.method static constructor <clinit>()V\n'
                             '   .locals 0\n'  # _revalue_locals_count() will ++ this
                             '\n'
                             '   .prologue\n'
                             '   const-string v0, "frida-gadget"\n'
                             '\n'
                             '   invoke-static {v0}, Ljava/lang/System;->loadLibrary(Ljava/lang/String;)V\n'
                             '\n'
                             '   return-void\n'
                             '.end method\n')

        # if an existing constructor is present, this partial_load_library
        # will be used instead
        partial_load_library = ('\n    const-string v0, "frida-gadget"\n'
                                '\n'
                                '    invoke-static {v0}, Ljava/lang/System;->loadLibrary(Ljava/lang/String;)V\n')

        # Check if there is an existing clinit here. If there is, then we need
        # to determine where the constructor ends and inject a simple loadLibrary
        # just before the end
        if 'clinit' in smali_lines[inject_marker]:
            click.secho('Injecting into an existing constructor', fg='yellow')

            inject_point = self._determine_first_inject_point_of_smali_method_from_line(smali_lines, inject_marker)
            click.secho('Injecting loadLibrary call at line: {0}'.format(inject_point), dim=True, fg='green')

            patched_smali = \
                smali_lines[:inject_point] + partial_load_library.splitlines(keepends=True) + \
                smali_lines[inject_point:]

        else:

            # if there is no constructor, we can simply inject a fresh constructor
            click.secho('Injecting loadLibrary call at line: {0}'.format(inject_marker), dim=True, fg='green')

            # inject the load_library code between
            patched_smali = \
                smali_lines[:inject_marker] + full_load_library.splitlines(keepends=True) + smali_lines[inject_marker:]

        return patched_smali

    def _revalue_locals_count(self, patched_smali: list, inject_marker: int):
        """
            Attempt to ++ the first .locals declaration in a list of
            smali lines confined to the same method.

            :param patched_smali:
            :param inject_marker:
            :return:
        """

        def _h():
            click.secho('Could not update .locals value. Sometimes this may break things,'
                        'but not always. If the applications crashes after patching, try '
                        'and add the --pause flag, fixing the patched smali manually.', fg='yellow')

        # next, update the .locals count (if its defined)
        # if this step fails, its not really a big deal as many times its not
        # fatal. however, if it does fail, warn about it.
        click.secho('Attempting to fix the constructors .locals count', dim=True)
        end_of_method = self._determine_end_of_smali_method_from_line(patched_smali, inject_marker)

        # check if we have a .locals declaration right after the start of our
        # already matched constructor
        defined_locals = [i for i, x in enumerate(patched_smali[inject_marker:end_of_method])
                          if '.locals' in x]

        if len(defined_locals) <= 0:
            click.secho('Unable to determine any .locals for the target constructor', fg='yellow')
            _h()
            return patched_smali

        # determine the offset for the first matched .locals definition
        locals_smali_offset = defined_locals[0] + inject_marker

        try:
            defined_local_value = patched_smali[locals_smali_offset].split(' ')[-1]
            defined_local_value_as_int = int(defined_local_value, 10)
            new_locals_value = defined_local_value_as_int + 1

        except ValueError as e:
            click.secho(
                'Unable to parse .locals value for the injected constructor with error: {0}'.format(str(e)),
                fg='yellow')
            _h()

            return patched_smali

        click.secho('Current locals value is {0}, updating to {1}:'.format(
            defined_local_value_as_int, new_locals_value), dim=True)

        # simply search / replace the integer values we already calculated on the relevant line
        patched_smali[locals_smali_offset] = patched_smali[locals_smali_offset].replace(
            str(defined_local_value_as_int), str(new_locals_value))

        return patched_smali

    def inject_load_library(self, target_class: str = None):
        """
            Injects a loadLibrary call into a class.
            If a target class is not specified, we will make an attempt
            at searching for a launchable activity in the target APK.

            Most of the idea for this comes from:
                https://koz.io/using-frida-on-android-without-root/

            :return:
        """

        # determine the path to the smali we should inject the load_library
        # call into. a user may specify a specific class to target, otherwise
        # we get a class name from the internal launchable activity method
        # of this class.

        if target_class:
            click.secho('Using target class: {0} for patch'.format(target_class), fg='green', bold=True)
        else:
            click.secho('Target class not specified, searching for launchable activity instead...', fg='green',
                        bold=True)

        activity_path = self._determine_smali_path_for_class(
            target_class if target_class else self._get_launchable_activity())

        click.secho('Reading smali from: {0}'.format(activity_path), dim=True)

        # apktool d smali will have a comment line line: '# direct methods'
        with open(activity_path, 'r') as f:
            smali_lines = f.readlines()

        # search for the line starting with '# direct methods' in it
        inject_marker = [i for i, x in enumerate(smali_lines) if '# direct methods' in x]

        # ensure we got a marker
        if len(inject_marker) <= 0:
            raise Exception('Unable to determine position to inject a loadLibrary call')

        # pick the first position for the inject. add one line as we
        # want to inject right below the comment we matched
        inject_marker = inject_marker[0] + 1

        patched_smali = self._patch_smali_with_load_library(smali_lines, inject_marker)
        patched_smali = self._revalue_locals_count(patched_smali, inject_marker)

        click.secho('Writing patched smali back to: {0}'.format(activity_path), dim=True)

        with open(activity_path, 'w') as f:
            f.write(''.join(patched_smali))

    def add_gadget_to_apk(self, architecture: str, gadget_source: str, gadget_config: str):
        """
            Copies a frida gadget for a specific architecture to
            an extracted APK's lib path.

            :param architecture:
            :param gadget_source:
            :param gadget_config:
            :return:
        """

        libs_path = os.path.join(self.apk_temp_directory, 'lib', architecture)

        # check if the libs path exists
        if not os.path.exists(libs_path):
            click.secho('Creating library path: {0}'.format(libs_path), dim=True)
            os.makedirs(libs_path)

        click.secho('Copying Frida gadget to libs path...', fg='green', dim=True)
        shutil.copyfile(gadget_source, os.path.join(libs_path, 'libfrida-gadget.so'))

        if gadget_config:
            click.secho('Adding a gadget configuration file...', fg='green')
            shutil.copyfile(gadget_config, os.path.join(libs_path, 'libfrida-gadget.config.so'))

    def build_new_apk(self, use_aapt2: bool = False, fix_concurrency_to = None):
        """
            Build a new .apk with the frida-gadget patched in.

            :return:
        """

        click.secho('Rebuilding the APK with the frida-gadget loaded...', fg='green', dim=True)
        o = delegator.run(
            self.list2cmdline([self.required_commands['apktool']['location'],
                            'build',
                            self.apk_temp_directory,
                            ] + (['--use-aapt2'] if use_aapt2 else []) + [
                                '-o',
                                self.apk_temp_frida_patched
                            ]+ ([] if fix_concurrency_to is None else ['-j', fix_concurrency_to]))
                            , timeout=self.command_run_timeout)
        

        if len(o.err) > 0:
            click.secho(('Rebuilding the APK may have failed. Read the following '
                         'output to determine if apktool actually had an error: \n'), fg='red')
            click.secho(o.err, fg='red')

        click.secho('Built new APK with injected loadLibrary and frida-gadget', fg='green')

    def zipalign_apk(self):
        """
            Performs the zipalign command on an APK.

            :return:
        """

        click.secho('Performing zipalign', dim=True)

        o = delegator.run(self.list2cmdline([
            self.required_commands['zipalign']['location'],
            '-p',
            '4',
            self.apk_temp_frida_patched if os.path.exists(self.apk_temp_frida_patched) else self.apk_source,
            self.apk_temp_frida_patched_aligned
        ]))

        if len(o.err) > 0:
            click.secho(('Zipaligning the APK may have failed. Read the following '
                         'output to determine if zipalign actually had an error: \n'), fg='red')
            click.secho(o.err, fg='red')

        click.secho('Zipalign completed', fg='green')

    def sign_apk(self):
        """
            Signs an APK with the objection key.

            The keystore itself was created with:
                keytool -genkey -v -keystore objection.jks -alias objection -keyalg RSA -keysize 2048 -validity 3650
                pass: basil-joule-bug

            :return:
        """

        click.secho('Signing new APK.', dim=True)

        o = delegator.run(self.list2cmdline([
            self.required_commands['apksigner']['location'],
            'sign',
            '--ks',
            self.keystore,
            '--ks-pass',
            'pass:basil-joule-bug',
            '--ks-key-alias',
            'objection',
            self.apk_temp_frida_patched_aligned
        ]))

        if len(o.err) > 0:
            click.secho('Signing the new APK may have failed.', fg='red')
            click.secho(o.out, fg='yellow')
            click.secho(o.err, fg='red')

        click.secho('Signed the new APK', fg='green')

    def __del__(self):
        """
            Cleanup after ourselves.

            :return:
        """

        if self.skip_cleanup:
            click.secho('Not cleaning up temporary files', dim=True)
            return

        click.secho('Cleaning up temp files...', dim=True)

        try:

            shutil.rmtree(self.apk_temp_directory, ignore_errors=True)

            with contextlib.suppress(FileNotFoundError):
                os.remove(self.apk_temp_frida_patched)

            with contextlib.suppress(FileNotFoundError):
                os.remove(self.apk_temp_frida_patched_aligned)

        except Exception as err:
            click.secho('Failed to cleanup with error: {0}'.format(err), fg='red', dim=True)



import logging
import os
import traceback

import click
import delegator
import frida
from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import FuzzyCompleter
from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.styles import Style

from .commands import COMMANDS
from .completer import CommandCompleter
from ..__init__ import __version__
from ..state.app import app_state
from ..state.connection import state_connection
from ..utils.helpers import get_tokens


class Repl(object):
    """
        The exploration REPL for objection
    """

    def __init__(self) -> None:
        self.cli = None

        self.completer = FuzzyCompleter(CommandCompleter())
        self.commands_repository = COMMANDS
        self.session = self.get_prompt_session()

    def get_prompt_session(self) -> PromptSession:
        """
            Starts a new prompt session.

            :return:
        """

        return PromptSession(
            history=FileHistory(os.path.expanduser('~/.objection/objection_history')),
            completer=self.completer,
            style=self.get_prompt_style(),
            auto_suggest=AutoSuggestFromHistory(),
            reserve_space_for_menu=4,
            complete_in_thread=True,
        )

    @staticmethod
    def get_prompt_style() -> Style:
        """
            Get the style to use for our prompt

            :return:
        """

        return Style.from_dict({
            # completions menu
            'completion-menu.completion.current': 'bg:#00aaaa #000000',
            'completion-menu.completion': 'bg:#008888 #ffffff',

            # fuzzy match outside
            'completion-menu.completion fuzzymatch.outside': 'fg:#000000',

            # Prompt.
            'applicationname': '#007cff',
            'status': '#717171',
            'on': '#00aa00',
            'devicetype': '#00ff48',
            'version': '#00ff48',
            'jobs': '',  # TODO
            'connection': '#717171'
        })

    @staticmethod
    def get_prompt_message() -> list:
        """
            Return prompt tokens to use in the cli app.

            If none were set during the init of this class, it
            is assumed that the connection failed.

            :return:
        """

        agent = state_connection.agent
        dev = state_connection.get_agent().device
        params = dev.query_system_parameters()

        return [
            ('class:applicationname', state_connection.name),
            ('class:status', ' (' + ('run' if agent.resumed else 'pause') + ')'),
            ('class:on', ' on '),
            ('class:devicetype', '(' + params['os']['name'] + ': '),
            ('class:version', params['os']['version'] + ') '),
            ('class:connection', '[' + dev.type + '] # '),
        ]

    def run_command(self, document: str) -> None:
        """
            Process a command as received by prompt_toolkit.

            :param document:
            :return:
        """

        logging.info(document)

        if document.strip() == '':
            return

        # handle os commands
        if document.strip().startswith('!'):

            # strip the leading !
            os_cmd = document[1:]

            click.secho('Running OS command: {0}\n'.format(os_cmd), dim=True)

            o = delegator.run(os_cmd, binary=True)

            # print stdout
            if len(o.out) > 0:
                click.secho(o.out.decode('utf-8', 'replace'), bold=True)

            # print stderr
            if len(o.err) > 0:
                click.secho(o.err.decode('utf-8', 'replace'), fg='red')

            return

        # a normal command is to be run, extract the tokens and
        # find which method we should be calling
        tokens = get_tokens(document)

        # check if we should be presenting help instead of executing
        # a command. this is indicated by the fact that the command
        # starts with the word 'help'
        if len(tokens) > 0 and 'help' == tokens[0]:

            # skip the 'help' entry from the tokens list so that
            # the following method can find the correct help
            tokens.remove('help')
            command_help = self._find_command_help(tokens)

            if not command_help:
                click.secho(('No help found for: {0}. Either the command '
                             'does not exist or contains subcommands with help.'
                             ).format(' '.join(tokens)), fg='yellow')
                return

            # output the help and leave
            click.secho(command_help, fg='blue', bold=True)
            return

        # find an execution method to run
        token_matches, exec_method = self._find_command_exec_method(tokens)

        if exec_method is None:
            click.secho('Unknown or ambiguous command: `{0}`. Try `help {0}`.'.format(document), fg='yellow')
            return

        # strip the command matching tokens and leave
        # the rest as arguments
        arguments = tokens[token_matches:]

        # run the method for the command itself!
        exec_method(arguments)

        app_state.add_command_to_history(command=document)

    def _find_command_exec_method(self, tokens: list) -> tuple:
        """
            Attempt to find the actual python method to run
            for the command tokens we have.

            This is done by 'walking' the command dictionary,
            looking for the deepest 'exec' method definition. We are
            interested in the number of tokens walked as well, so
            that the calling command can know how many tokens to
            strip, sending the rest as arguments to the exec method.

            :param tokens:
            :return:
        """

        # start with all of the commands we have
        dict_to_walk = self.commands_repository

        # ... and an empty method to execute
        exec_method = None

        # keep count of the number of tokens
        # used in this walk. this will help indicate to
        # the caller how many tokens should be stripped to
        # get to the arguments for the command
        walked_tokens = 0

        for token in tokens:

            # increment the walked tokens
            walked_tokens += 1

            # check if the token matches a command
            if token in dict_to_walk:

                # matched a dict for the token we have. we need
                # to have *all* of the tokens match a nested dict
                # so that we can extract the final 'exec' key.
                # if we encounter a key that does not have nested commands,
                # chances are we are where we need to be to exec a command.
                if 'commands' not in dict_to_walk[token]:

                    if 'exec' in dict_to_walk[token]:
                        exec_method = dict_to_walk[token]['exec']
                        break

                else:
                    dict_to_walk = dict_to_walk[token]['commands']

            # stop if there is nothing that matches
            else:
                break

        return walked_tokens, exec_method

    def _find_command_help(self, tokens: list) -> str:
        """
            Attempt to find help for a command.

            Just like how the _find_command_exec_method works, this
            method also walks the command dictionary, searching for
            the deepest key. The tokens that match form part of a
            new list, later joined together to pickup the correct
            help.txt.

            :param tokens:
            :return:
        """

        # start with all of the commands we have
        dict_to_walk = self.commands_repository
        helpfile_name = []
        user_help = ''

        for token in tokens:

            # check if the token matches a command
            if token in dict_to_walk:

                # add this token to the helpfile
                helpfile_name.append(token)

                # if there are subcommands, continue with the walk
                if 'commands' in dict_to_walk[token]:
                    dict_to_walk = dict_to_walk[token]['commands']

            # stop if we don't have a token that matches anything
            else:
                break

        # once we have the help, load its .txt contents
        if len(helpfile_name) > 0:

            help_file = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                     'helpfiles', '.'.join(helpfile_name) + '.txt')

            # no helpfile... warn.
            if not os.path.exists(help_file):
                click.secho('Unable to find helpfile {0}'.format(' '.join(helpfile_name)), dim=True)

                return user_help

            # read the helpfile
            with open(help_file, 'r') as f:
                user_help = f.read()

        return user_help

    @staticmethod
    def handle_reconnect(document: str) -> bool:
        """
            Handles a reconnection attempt to a device.

            A reconnection means that the current agent will be unloaded
            and reloaded again.

            :param document:
            :return:
        """

        if document.strip() in ('reconnect', 'reset', 'reconnect_spawn'):
            try:
                from .cli import get_agent

                reconnect_spawn = document.strip() == 'reconnect_spawn'
                if reconnect_spawn:
                    click.secho('Performing full-restart...', fg='yellow')
                    state_connection.spawn = True
                    state_connection.no_pause = True
                else:
                    click.secho('Performing soft-restart...', fg='yellow')
                    state_connection.spawn = False

                curr_agent = state_connection.agent

                # Cleanup current agent (ignore errors if already destroyed)
                click.secho('Unloading current agent...', dim=True)
                try:
                    if curr_agent.script:
                        curr_agent.script.unload()

                except (frida.InvalidOperationError, Exception):
                    pass  # Script already destroyed or detached

                try:
                    if curr_agent.session:
                        curr_agent.session.detach()
                except (frida.InvalidOperationError, Exception):
                    pass  # Session already detached
                
                # Need to clear because destructor will attempt to clear script/session again.
                curr_agent.script = None
                state_connection.agent = None
                state_connection.session = None

                click.secho(f'Re-attaching to {state_connection.name}...', dim=True)

                # Try respawn the agent.
                new_agent = get_agent()
                state_connection.agent = new_agent
                
                click.secho('Reconnection successful!', fg='green')

            except Exception as e:
                click.secho(f'Reconnection failed: {e}', fg='red')
                click.secho('Ensure the application is running and the device is connected.', dim=True)

            return True

        return False

    def run(self, quiet: bool) -> None:
        """
            Start the objection repl.
        """

        banner = ("""
     _   _         _   _
 ___| |_|_|___ ___| |_|_|___ ___
| . | . | | -_|  _|  _| | . |   |
|___|___| |___|___|_| |_|___|_|_|
      |___|(object)inject(ion) v{0}

     Runtime Mobile Exploration
        by: @leonjza from @sensepost
""").format(__version__)

        if not quiet:
            click.secho(banner, bold=True)
            click.secho('[tab] for command suggestions', fg='white', dim=True)

        # the main application loop is here, reading inputs provided by
        # prompt_toolkit and sending it off the needed handlers
        while True:

            try:
                with patch_stdout(raw=True):
                    document = self.session.prompt(self.get_prompt_message())

                    # check if this is an exit command
                    if document.strip() in ('quit', 'exit', 'bye'):
                        click.secho('Exiting...', dim=True)
                        break

                    if document.strip() in ('resume', 'res'):
                        click.secho('Resuming attached process', dim=True)
                        state_connection.agent.resume()
                        continue

                    # if we got the reconnect command, handle just that
                    if self.handle_reconnect(document):
                        continue

                    # dispatch to the command handler. if something goes horribly
                    # wrong, catch it instead of crashing the REPL
                    try:

                        # find something to run
                        self.run_command(document)

                    except frida.core.RPCException as e:
                        click.secho('A Frida agent exception has occurred.', fg='red', bold=True)
                        click.secho('{0}'.format(e), fg='red')
                        click.secho('\nPython stack trace: {}'.format(traceback.format_exc()), dim=True)

                    except Exception as e:
                        click.secho(('An unexpected internal exception has occurred. If this '
                                     'looks like a code related error, please file a bug report!'), fg='red', bold=True)
                        click.secho('{0}'.format(e), fg='red')
                        click.secho('\nPython stack trace: {}'.format(traceback.format_exc()), dim=True)

            except KeyboardInterrupt:
                pass

            except EOFError:
                click.secho('Exiting...', dim=True)
                break



import os

import click

from objection.commands import filemanager
from objection.state.connection import state_connection
from objection.state.device import device_state


def cat(args: list = None) -> None:
    """
        Parses a plist on an iOS device and echoes it in a more human
        readable way.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: ios plist cat <remote_plist>', bold=True)
        return

    plist = args[0]

    if not os.path.isabs(plist):
        pwd = filemanager.pwd()
        plist = device_state.platform.path_separator.join([pwd, plist])

    api = state_connection.get_api()
    plist_data = api.ios_plist_read(plist)

    click.secho(plist_data, bold=True)



import os

import click
import frida
from prompt_toolkit import prompt
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.javascript import JavascriptLexer

from ..state.connection import state_connection


def evaluate(args: list) -> None:
    """
        Evaluate JavaScript within the agent's context.

        :param args:
        :return:
    """

    target_file = None

    # if we have an argument, let's assume it is a file path
    if len(args) > 0:

        target_file = args[0]
        p = os.path.expanduser(target_file)
        if os.path.exists(p):
            target_file = p
        else:
            click.secho('Could not find file {p}.'.format(p=target_file), fg='red')
            return

    if target_file:
        with open(target_file, 'r', encoding='utf-8') as f:
            javascript = ''.join(f.readlines())
    else:
        javascript = prompt(
            multiline=True, lexer=PygmentsLexer(JavascriptLexer),
            bottom_toolbar='JavaScript edit mode. [ESC] and then [ENTER] to accept. [CTRL] + C to cancel.').strip()

    if len(javascript) <= 0:
        click.secho('JavaScript to evaluate appears empty. Skipping.', fg='yellow')
        return

    click.secho('JavaScript capture complete. Evaluating...', dim=True)
    try:
        state_connection.get_api().evaluate(javascript)
    except frida.core.RPCException as e:
        click.secho('Failed to load script: {}'.format(e), fg='red', bold=True)



from objection.state.connection import state_connection


def deoptimise(args: list) -> None:
    """
        Forces the VM to execute everything with its interpreter.
        Necessary to prevent optimizations from bypassing method hooks in some cases.

        Ref: https://frida.re/docs/javascript-api/

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.android_deoptimize()



from ..api.app import create_app


class ApiState(object):
    """ A class representing the state API for this app """

    def __init__(self):
        self.core_api = create_app()
        self.blueprints = []

    def append_api_blueprint(self, blueprint):
        """
            Add extra blueprints to the API.

            This method would typically be called by the
            plugin loader to slot in endpoints that plugins
            may expose.

            :param blueprint:
            :return:
        """

        self.blueprints.append(blueprint)

    def start(self, host: str, port: int, debug: bool = False):
        """
            Starts the Flask-based API server after
            registering any extra blueprints that would
            typically have been sources from plugins.

            :param host:
            :param port:
            :param debug:
            :return:
        """

        for bp in self.blueprints:
            self.core_api.register_blueprint(bp)

        self.core_api.run(host=host, port=port, debug=debug)


api_state = ApiState()



from objection.state.connection import state_connection


def crypto_enable(args: list = None) -> None:
    """
        Attempts to enable ios crypto monitoring.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.ios_monitor_crypto_enable()



import json
import os
from datetime import datetime, timedelta

import click
import requests
from packaging.version import Version

from ..__init__ import __version__

objection_path = os.path.join(os.path.expanduser('~'), '.objection')
version_file = os.path.join(objection_path, 'version_info')

version_data = {
    'remote_version': '0.0.0',
    'last_check': datetime.now() - timedelta(days=7)
}

date_fmt = '%d%m%y %H:%M:%S'


def cached_version_data() -> version_data:
    """
        Reads the local version file and returns the
        version data structure

        :return:version_data
    """

    if not os.path.exists(version_file):
        return version_data

    with open(version_file, 'r') as f:
        data = json.load(f)

    data['last_check'] = datetime.strptime(data['last_check'], date_fmt)

    return data


def update_version_cache(version: str) -> None:
    """
        Store version information.

        :param version:
        :return:
    """

    version_data['remote_version'] = version
    version_data['last_check'] = datetime.now().strftime(date_fmt)

    with open(version_file, 'w') as f:
        json.dump(version_data, f)


def notify_newer_version() -> None:
    """
        Print a notification message about the newer version
        that is available.

        :return:
    """

    cache_version = cached_version_data()['remote_version']

    if Version(cache_version) > Version(__version__):
        click.secho('\n\nA newer version of objection is available!', fg='green')
        click.secho('You have v{0} and v{1} is ready for download.\n'.format(
            __version__, cache_version), fg='green')
        click.secho('Upgrade with: pip3 install objection --upgrade', fg='green', bold=True)
        click.secho('For more information, please see: '
                    'https://github.com/sensepost/objection/wiki/Updating\n', dim=True)


def check_version() -> None:
    """
        Checks if the current version of objection is up to date.

        :return:
    """

    # if we have not checked for a new version today
    if not (cached_version_data()['last_check'] > datetime.now() - timedelta(hours=23)):
        click.secho('Checking for a newer version of objection...', dim=True)

        # noinspection PyBroadException
        try:

            r = requests.get('https://api.github.com/repos/sensepost/objection/releases/latest').json()
            update_version_cache(r['tag_name'])

        # Just be quiet about any exceptions here. If this method fails
        # it really doesn't matter.
        except Exception:
            # there is good chance an installation does not have internet, so cache
            # the current version as the latest to not be annoying about updates
            update_version_cache(__version__)

    notify_newer_version()



import json
from typing import Optional

import click

from objection.state.connection import state_connection
from objection.utils.helpers import clean_argument_flags

# a thumb sucked list of prefixes used in Objective-C runtime
# for iOS applications. This is not a science, but a gut feeling.
native_prefixes = [
    '_',
    'NS',
    # '_NS',
    # '__NS',
    'CF',
    'OS_',
    'UI',
    # '_UI',

    'AWD',
    'GEO',

    'AC',
    'AF',
    'AU',
    'AV',
    'BK',
    'BS',
    'CA',
    'CB',
    'CI',
    'CL',
    'CT',
    'CUI',
    'DOM',
    'FBS',
    'LA',
    'LS',
    'MC',
    'MTL',
    'PFUbiquity',
    'PKPhysics',
    'SBS',
    'TI',
    'TXR',
    'UM',
    'Web',
]


def _should_ignore_native_classes(args: list) -> bool:
    """
        Checks if --ignore-native is in a list of tokens received
        from the commandline.

        :param args:
        :return:
    """

    if len(args) <= 0:
        return False

    return '--ignore-native' in args


def _should_include_parent_methods(args: list) -> bool:
    """
        Checks if --include-parents exists in a list of tokens received
        from the commandline.

        :param args:
        :return:
    """

    if len(args) <= 0:
        return False

    return '--include-parents' in args


def _class_is_prefixed_with_native(class_name: str) -> bool:
    """
        Check if a class name received is prefixed with one of the
        prefixes in the native_prefixes list.

        :param class_name:
        :return:
    """

    for prefix in native_prefixes:

        if class_name.startswith(prefix):
            return True

    return False


def _string_is_true(s: str) -> bool:
    """
        Check if a string should be considered as "True"

        :param s:
        :return:
    """

    return s.lower() in ('true', 'yes')


def _should_dump_backtrace(args: list) -> bool:
    """
        Check if --dump-backtrace is part of the arguments.

        :param args:
        :return:
    """

    return '--dump-backtrace' in args


def _should_dump_args(args: list) -> bool:
    """
        Check if --dump-args is part of the arguments.

        :param args:
        :return:
    """

    return '--dump-args' in args


def _should_dump_return_value(args: list) -> bool:
    """
        Check if --dump-return is part of the arguments.

        :param args:
        :return:
    """

    return '--dump-return' in args


def _should_print_only_classes(args: list) -> bool:
    """
        Check if --only-classes is part of the arguments.

        :param args:
        :return:
    """

    return '--only-classes' in args


def _should_dump_json(args: list) -> bool:
    """
        Check if --json is part of the arguments.

        :param args:
        :return:
    """

    return '--json' in args


def _should_be_quiet(args: list) -> bool:
    """
        Check if --quiet is part of the arguments.

        :param args:
        :return:
    """

    return '--quiet' in args


def _get_flag_value(flag: str, args: list) -> Optional[str]:
    """
        Gets the value for a flag

        :param flag:
        :param args:
        :return:
    """

    target = None

    for i in range(len(args)):
        if args[i] == flag:
            target = i + 1

    if target is None:
        return None
    elif target < len(args):
        return args[target]
    else:
        return None


def show_ios_classes(args: list = None) -> None:
    """
        Prints the classes available in the current Objective-C
        runtime to the screen.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    classes = api.ios_hooking_get_classes()

    # loop the class names and check if we should be ignoring it.
    for class_name in sorted(classes):
        if _should_ignore_native_classes(args):
            if not _class_is_prefixed_with_native(class_name):
                click.secho(class_name)
                continue

        else:
            click.secho(class_name)

    click.secho('\nFound {0} classes'.format(len(classes)), bold=True)


def show_ios_class_methods(args: list) -> None:
    """
        Displays the methods available in a class.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: ios hooking list class_methods <class name> (--include-parents)', bold=True)
        return

    classname = args[0]

    api = state_connection.get_api()
    methods = api.ios_hooking_get_class_methods(classname, _should_include_parent_methods(args))

    if len(methods) > 0:

        # dump the methods to screen
        for method in methods:
            click.secho(method)

        click.secho('\nFound {0} methods'.format(len(methods)), bold=True)

    else:
        click.secho('No class / methods found')


def set_method_return_value(args: list) -> None:
    """
        Make an Objective-C method return a specific boolean
        value, always.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) < 2:
        click.secho('Usage: ios hooking set_method_return "<selector>" (eg: "-[ClassName methodName:]") <true/false>',
                    bold=True)
        return

    selector = args[0]
    retval = args[1]

    api = state_connection.get_api()
    api.ios_hooking_set_return_value(selector, _string_is_true(retval))


def watch(args: list) -> None:
    """
        Watches a pattern for invocations.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: ios hooking watch <pattern>', bold=True)
        return

    pattern = args[0]

    api = state_connection.get_api()
    api.ios_hooking_watch(pattern,
                          _should_dump_args(args),
                          _should_dump_backtrace(args),
                          _should_dump_return_value(args),
                          _should_include_parent_methods(args))


def search(args: list) -> None:
    """
        Searches the current iOS application for classes and methods.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: ios hooking search \'<pattern/string>\'', bold=True)
        return

    api = state_connection.get_api()
    pattern = args[0]

    results = api.ios_hooking_search(pattern)
    data = {}

    # build a list of results to print / dump later
    for func in results:
        fullname = func['name']
        start_bracket = fullname.find('[') + 1
        class_name = fullname[start_bracket: fullname.find(' ')]
        if data.get(class_name) is not None:
            data[class_name].append(fullname)
        else:
            data[class_name] = [fullname]

    if _should_dump_json(args):
        target_file = _get_flag_value('--json', args)
        if not target_file:
            click.secho('A file name needs to be specified with the --json flag', fg='red')
            return

        with open(target_file, 'w') as fd:
            fd.write(json.dumps({
                'meta': {
                    'runtime': 'objc'
                },
                'classes': data
            }))
            click.secho(f'JSON dumped to file {target_file}', bold=True)

        return

    # Print the matching methods
    for klass in data.keys():
        if _should_print_only_classes(args):
            print(klass)
            continue

        methods = data[klass]
        for method in methods:
            print(f'{method}')



from objection.commands import http
from ..commands import command_history
from ..commands import custom
from ..commands import device
from ..commands import filemanager
from ..commands import frida_commands
from ..commands import jobs
from ..commands import memory
from ..commands import plugin_manager
from ..commands import sqlite
from ..commands import ui
from ..commands.android import clipboard
from ..commands.android import command
from ..commands.android import general
from ..commands.android import generate as android_generate
from ..commands.android import heap as android_heap
from ..commands.android import hooking as android_hooking
from ..commands.android import intents
from ..commands.android import keystore
from ..commands.android import pinning as android_pinning
from ..commands.android import proxy as android_proxy
from ..commands.android import root
from ..commands.ios import binary
from ..commands.ios import bundles
from ..commands.ios import cookies
from ..commands.ios import generate as ios_generate
from ..commands.ios import heap as ios_heap
from ..commands.ios import hooking as ios_hooking
from ..commands.ios import jailbreak
from ..commands.ios import keychain
from ..commands.ios import monitor as ios_crypto
from ..commands.ios import nsurlcredentialstorage
from ..commands.ios import nsuserdefaults
from ..commands.ios import pasteboard
from ..commands.ios import pinning as ios_pinning
from ..commands.ios import plist

# commands are defined with their name being the key, then optionally
# have a meta, dynamic and commands key.

# meta: A small one-liner containing information about the command itself
# dynamic: A method to execute that would return completions to populate in the prompt
# exec: The *actual* method to execute when the command is issued.

# commands help is stored in the help files directory as a txt file.

COMMANDS = {

    'plugin': {
        'meta': 'Work with plugins',
        'commands': {
            'load': {
                'meta': 'Load a plugin',
                'exec': plugin_manager.load_plugin
            }
        }
    },

    '!': {
        'meta': 'Execute an Operating System command',
        'exec': None,  # handled in the Repl class itself
    },

    'reconnect': {
        'meta': 'Reconnect to the current app',
        'exec': None,  # handled in the Repl class itself
    },

    'reconnect_spawn': {
        'meta': 'Respawn the current app',
        'exec': None,  # handled in the Repl class itself
    },

    'resume': {
        'meta': 'Resume the attached process',
        'exec': None
    },

    'import': {
        'meta': 'Import fridascript from a full path and run it',
        'exec': frida_commands.load_background
    },

    'ping': {
        'meta': 'Ping the injected agent',
        'exec': frida_commands.ping
    },

    # file manager commands

    'cd': {
        'meta': 'Change the current working directory',
        'dynamic': filemanager.list_folders_in_current_fm_directory,
        'exec': filemanager.cd
    },

    'commands': {
        'meta': 'Work with commands run in the current session',
        'commands': {
            'history': {
                'meta': 'List all unique commands that have run for this session',
                'exec': command_history.history,
            },
            'save': {
                'meta': 'Save all unique commands that have run in this session to a file',
                'exec': command_history.save
            },
            'clear': {
                'meta': 'Clear the current sessions command history',
                'exec': command_history.clear
            }
        }
    },

    'ls': {
        'meta': 'List files in the current working directory',
        'dynamic': filemanager.list_folders_in_current_fm_directory,
        'exec': filemanager.ls,
    },

    'pwd': {
        'meta': 'Print the current working directory on the device',
        'exec': filemanager.pwd_print,
    },

    'filesystem': {
        'meta': 'Work with files on the remote filesystem',
        'commands': {
            'cat': {
                'meta': 'Print a files contents',
                'dynamic': filemanager.list_files_in_current_fm_directory,
                'exec': filemanager.cat
            },
            'upload': {
                'meta': 'Upload a file',
                'exec': filemanager.upload
            },
            'download': {
                'meta': 'Download a file or folder',
                'flags': ['--folder'],
                'dynamic': filemanager.list_content_in_current_fm_directory,
                'exec': filemanager.download
            },

            # http file server

            # 'http': {
            #     'meta': 'Work with an on device HTTP file server',
            #     'commands': {
            #         'start': {
            #             'meta': 'Start\'s an HTTP server in the current working directory',
            #             'exec': http.start
            #         },
            #         'status': {
            #             'meta': 'Get the status of the HTTP server',
            #             'exec': http.status
            #         },
            #         'stop': {
            #             'meta': 'Stop\'s a running HTTP server',
            #             'exec': http.stop
            #         }
            #     }
            # },
        }
    },

    'rm': {
        'meta': 'Delete files from the remote filesystem',
        'dynamic': filemanager.list_files_in_current_fm_directory,
        'exec': filemanager.rm
    },

    # device and env info commands

    'env': {
        'meta': 'Print information about the environment',
        'exec': device.get_environment
    },

    'frida': {
        'meta': 'Get information about the Frida environment',
        'exec': frida_commands.frida_environment
    },

    'evaluate': {
        'meta': 'Evaluate JavaScript within the agent',
        'exec': custom.evaluate
    },

    # memory commands

    'memory': {
        'meta': 'Work with the current processes memory',
        'commands': {
            'dump': {
                'meta': 'Commands to dump parts of the processes memory',
                'commands': {
                    'all': {
                        'meta': 'Dump the entire memory of the current process',
                        'exec': memory.dump_all
                    },

                    'from_base': {
                        'meta': 'Dump (x) bytes of memory from a base address to file',
                        'exec': memory.dump_from_base
                    }
                },
            },

            'list': {
                'meta': 'List memory related information about the current process',
                'commands': {
                    'modules': {
                        'meta': 'List loaded modules in the current process',
                        'flags': ['--json'],
                        'exec': memory.list_modules
                    },

                    'exports': {
                        'meta': 'List the exports of a module',
                        'flags': ['--json'],
                        'exec': memory.list_exports
                    }
                },
            },

            'search': {
                'meta': 'Search for pattern in the applications memory',
                'flags': ['--string', '--offsets-only'],
                'exec': memory.find_pattern
            },

            'replace': {
                'meta': 'Search and replace pattern in the applications memory',
                'flags': ['--string-pattern', '--string-replace'],
                'exec': memory.replace_pattern
            },

            'write': {
                'meta': 'Write raw bytes to a memory address. Use with caution!',
                'flags': ['--string'],
                'exec': memory.write
            }
        },
    },

    # sqlite commands

    'sqlite': {
        'meta': 'Work with SQLite databases',
        'commands': {
            'connect': {
                'meta': 'Connect to a SQLite database file',
                'flags': ['--sync'],
                'dynamic': filemanager.list_files_in_current_fm_directory,
                'exec': sqlite.connect
            },
        }
    },

    # jobs commands

    'jobs': {
        'meta': 'Work with objection jobs',
        'commands': {
            'list': {
                'meta': 'List all of the current jobs',
                'exec': jobs.show
            },
            'kill': {
                'meta': 'Kill a job. This unloads the script',
                'dynamic': jobs.list_current_jobs,
                'exec': jobs.kill
            }
        }
    },

    # generic ui commands

    'ui': {
        'meta': 'Generic user interface commands',
        'commands': {
            'alert': {
                'meta': 'Show an alert message, optionally specifying the message to show. (Currently crashes iOS)',
                'exec': ui.alert
            }
        }
    },

    # android commands

    'android': {
        'meta': 'Commands specific to Android',
        'commands': {
            'deoptimize': {
                'meta': 'Force the VM to execute everything in the interpreter',
                'exec': general.deoptimise
            },
            'shell_exec': {
                'meta': 'Execute a shell command',
                'exec': command.execute
            },
            'hooking': {
                'meta': 'Commands used for hooking methods in Android',
                'commands': {
                    'list': {
                        'meta': 'Lists various bits of information',
                        'commands': {
                            'classes': {
                                'meta': 'List the currently loaded classes',
                                'exec': android_hooking.show_android_classes
                            },
                            'class_methods': {
                                'meta': 'List the methods available on a class',
                                'exec': android_hooking.show_android_class_methods
                            },
                            'class_loaders': {
                                'meta': 'List the registered class loaders',
                                'exec': android_hooking.show_android_class_loaders
                            },
                            'activities': {
                                'meta': 'List the registered Activities',
                                'exec': android_hooking.show_registered_activities
                            },
                            'receivers': {
                                'meta': 'List the registered BroadcastReceivers',
                                'exec': android_hooking.show_registered_broadcast_receivers
                            },
                            'services': {
                                'meta': 'List the registered Services',
                                'exec': android_hooking.show_registered_services
                            },
                        }
                    },
                    'watch': {
                        'meta': 'Watch for Android Java invocations',
                        'exec': android_hooking.watch,
                        'flags': ['--dump-args', '--dump-backtrace', '--dump-return']
                    },
                    'set': {
                        'meta': 'Set various values',
                        'commands': {
                            'return_value': {
                                'meta': 'Set a methods return value. Supports only boolean returns.',
                                'exec': android_hooking.set_method_return_value,
                                'flags': ['--dump-args', '--dump-return', '--dump-backtrace']
                            }
                        }
                    },
                    'get': {
                        'meta': 'Get various values',
                        'commands': {
                            'current_activity': {
                                'meta': 'Get the currently foregrounded activity',
                                'exec': android_hooking.get_current_activity
                            }
                        }
                    },
                    'search': {
                        'meta': 'Search for various classes and or methods',
                        'exec': android_hooking.search,
                        'flags': ['--json', '--only-classes']
                    },
                    'notify': {
                        'meta': 'Notify when a class becomes available',
                        'exec': android_hooking.notify,
                        'flags': ['--dump-args', '--dump-return', '--dump-backtrace', '--watch']

                    },
                    'generate': {
                        'meta': 'Generate Frida hooks for Android',
                        'commands': {
                            'class': {
                                'meta': 'A generic hook manager for Classes',
                                'exec': android_generate.clazz
                            },
                            'simple': {
                                'meta': 'Simple hooks for each Class method',
                                'exec': android_generate.simple
                            }
                        }
                    }
                },
            },
            'heap': {
                'meta': 'Commands to work with the Android Heap',
                'commands': {
                    'search': {
                        'meta': 'Search for information about the current Android heap',
                        'commands': {
                            'instances': {
                                'meta': 'Search for live instances of a particular class',
                                'exec': android_heap.instances

                            }
                        }
                    },
                    'print': {
                        'meta': 'Print information about objects on the Android heap',
                        'commands': {
                            'fields': {
                                'meta': 'Print instance fields for a Java object handle',
                                'exec': android_heap.fields
                            },
                            'methods': {
                                'meta': 'Print instance methods for an Android handle',
                                'flags': ['--without-arguments'],
                                'exec': android_heap.methods
                            }
                        }
                    },
                    'execute': {
                        'meta': 'Execute methods on Java class handles',
                        'flags': ['--return-string'],
                        'exec': android_heap.execute
                    },
                    'evaluate': {
                        'meta': 'Evaluate JavaScript on Java class handle',
                        'exec': android_heap.evaluate
                    }
                }
            },
            'keystore': {
                'meta': 'Commands to work with the Android KeyStore',
                'commands': {
                    'list': {
                        'meta': 'Lists entries in the Android KeyStore',
                        'exec': keystore.entries
                    },
                    'detail': {
                        'meta': 'Lists details for all items in the Android KeyStore',
                        'flags': ['--json'],
                        'exec': keystore.detail
                    },
                    'clear': {
                        'meta': 'Clears the Android KeyStore',
                        'exec': keystore.clear
                    },
                    'watch': {
                        'meta': 'Watches usage of the Android keystore',
                        'exec': keystore.watch
                    }
                }
            },
            'clipboard': {
                'meta': 'Work with the Android Clipboard',
                'commands': {
                    'monitor': {
                        'meta': 'Monitor the Android Clipboard',
                        'exec': clipboard.monitor
                    }
                }
            },
            'intent': {
                'meta': 'Commands to work with Android intents',
                'commands': {
                    'launch_activity': {
                        'meta': 'Launch an Activity class using an Intent',
                        'exec': intents.launch_activity
                    },
                    'launch_service': {
                        'meta': 'Launch a Service class using an Intent',
                        'exec': intents.launch_service
                    },
                    'implicit_intents': {
                        'meta': 'Analyze implicit intents',
                        'exec': intents.analyze_implicit_intents,
                        'flags': ['--dump-backtrace']
                    }
                }
            },
            'root': {
                'meta': 'Commands to work with Android root detection',
                'commands': {
                    'disable': {
                        'meta': 'Attempt to disable root detection',
                        'exec': root.disable
                    },
                    'simulate': {
                        'meta': 'Attempt to simulate a rooted environment',
                        'exec': root.simulate
                    }
                }
            },
            'sslpinning': {
                'meta': 'Work with Android SSL pinning',
                'commands': {
                    'disable': {
                        'meta': 'Attempt to disable SSL pinning in various Java libraries/classes',
                        'flags': ['--quiet'],
                        'exec': android_pinning.android_disable
                    }
                }
            },
            'proxy': {
                'meta': 'Commands to work with a proxy for the application',
                'commands': {
                    'set': {
                        'meta': 'Set a proxy for the application',
                        'exec': android_proxy.android_proxy_set
                    }
                }
            },
            'ui': {
                'meta': 'Android user interface commands',
                'commands': {
                    'screenshot': {
                        'meta': 'Screenshot the current Activity',
                        'exec': ui.android_screenshot
                    },
                    'FLAG_SECURE': {
                        'meta': 'Control FLAG_SECURE of the current Activity',
                        'exec': ui.android_flag_secure
                    },
                }
            },
        },
    },
    # ios commands
    'ios': {
        'meta': 'Commands specific to iOS',
        'commands': {
            'info': {
                'meta': 'Get iOS and application related information',
                'commands': {
                    'binary': {
                        'meta': 'Get information about application binaries and dylibs',
                        'exec': binary.info
                    }
                }
            },
            'keychain': {
                'meta': 'Work with the iOS keychain',
                'commands': {
                    'dump': {
                        'meta': 'Dump the keychain for the current app\'s entitlement group',
                        'flags': ['--json', '--smart'],
                        'exec': keychain.dump
                    },
                    'dump_raw': {
                        'meta': 'Dump raw, unprocessed keychain entries (advanced)',
                        'exec': keychain.dump_raw
                    },
                    'clear': {
                        'meta': 'Delete all keychain entries for the current app\'s entitlement group',
                        'exec': keychain.clear
                    },
                    'remove': {
                        'meta': 'Remove an entry from the iOS keychain',
                        'flags': ['--account', '--service'],
                        'exec': keychain.remove
                    },
                    'update': {
                        'meta': 'Update an entry from the iOS keychain',
                        'flags': ['--account', '--service', '--newData'],
                        'exec': keychain.update
                    },                     
                    'add': {
                        'meta': 'Add an entry to the iOS keychain',
                        'flags': ['--account', '--service', '--data'],
                        'exec': keychain.add
                    }
                }
            },
            'plist': {
                'meta': 'Work with iOS Plists',
                'commands': {
                    'cat': {
                        'meta': 'Cat a plist',
                        'dynamic': filemanager.list_files_in_current_fm_directory,
                        'exec': plist.cat
                    }
                }
            },
            'bundles': {
                'meta': 'Work with iOS Bundles',
                'commands': {
                    'list_frameworks': {
                        'meta': 'Lists all of the application\'s bundles that represent frameworks',
                        'flags': ['--include-apple-frameworks', '--full-path'],
                        'exec': bundles.show_frameworks
                    },
                    'list_bundles': {
                        'meta': 'Lists all of the application\'s non framework bundles',
                        'flags': ['--full-path'],
                        'exec': bundles.show_bundles
                    }
                }
            },
            'nsuserdefaults': {
                'meta': 'Work with NSUserDefaults',
                'commands': {
                    'get': {
                        'meta': 'Get all of the entries',
                        'exec': nsuserdefaults.get
                    }
                }
            },
            'nsurlcredentialstorage': {
                'meta': 'Work with the shared NSURLCredentialStorage',
                'commands': {
                    'dump': {
                        'meta': 'Dump all of the credentials in the shared NSURLCredentialStorage',
                        'exec': nsurlcredentialstorage.dump
                    }
                }
            },
            'cookies': {
                'meta': 'Work with shared cookies',
                'commands': {
                    'get': {
                        'meta': 'Get the current apps shared cookies',
                        'flags': ['--json'],
                        'exec': cookies.get
                    }
                }
            },
            'ui': {
                'meta': 'iOS user interface commands',
                'commands': {
                    'alert': {
                        'meta': ('Show an alert message, optionally specifying the message to'
                                 'show. (Currently crashes iOS)'),
                        'exec': ui.alert
                    },
                    'dump': {
                        'meta': 'Dump the serialized UI',
                        'exec': ui.dump_ios_ui
                    },
                    'screenshot': {
                        'meta': 'Screenshot the current UIView',
                        'exec': ui.ios_screenshot
                    },
                    'biometrics_bypass': {
                        'meta': 'Hook the iOS Biometrics LAContext and respond with successful auth',
                        'exec': ui.bypass_touchid
                    }
                }
            },
            'heap': {
                'meta': 'Commands to work with the iOS heap',
                'commands': {
                    'print': {
                        'meta': 'Print information about objects on the iOS heap',
                        'commands': {
                            'ivars': {
                                'meta': 'Print instance variables for an Objective-C object',
                                'flags': ['--to-utf8'],
                                'exec': ios_heap.ivars
                            },
                            'methods': {
                                'meta': 'Print instance methods for an Objective-C object',
                                'flags': ['--without-arguments'],
                                'exec': ios_heap.methods
                            }
                        }
                    },
                    'search': {
                        'meta': 'Search for information about the current iOS heap',
                        'commands': {
                            'instances': {
                                'meta': 'Search for live instances of a particular class',
                                'exec': ios_heap.instances
                            }
                        }
                    },
                    'execute': {
                        'meta': 'Execute methods on objects on the iOS heap',
                        'flags': ['--return-string'],
                        'exec': ios_heap.execute
                    },
                    'evaluate': {
                        'meta': 'Evaluate JavaScript on objects on the iOS heap',
                        'flags': ['--inline'],
                        'exec': ios_heap.evaluate
                    }
                }
            },
            'hooking': {
                'meta': 'Commands used for hooking methods in iOS',
                'commands': {
                    'list': {
                        'meta': 'Lists various bits of information',
                        'commands': {
                            'classes': {
                                'meta': 'List classes available in the current application',
                                'exec': ios_hooking.show_ios_classes
                            },
                            'class_methods': {
                                'meta': 'List the methods in a class',
                                'flags': ['--include-parents'],
                                'exec': ios_hooking.show_ios_class_methods
                            }
                        }
                    },
                    'watch': {
                        'meta': 'Watch invocations of classes and methods',
                        'exec': ios_hooking.watch,
                        'flags': ['--dump-args', '--dump-backtrace', '--dump-return'],
                    },
                    'set': {
                        'meta': 'Set various values',
                        'commands': {
                            'return_value': {
                                'meta': 'Set a methods return value. Supports only boolean returns',
                                'exec': ios_hooking.set_method_return_value
                            }
                        }
                    },
                    'search': {
                        'meta': 'Search for various classes and or methods',
                        'exec': ios_hooking.search,
                        'flags': ['--json', '--only-classes']
                    },
                    'generate': {
                        'meta': 'Generate Frida hooks for iOS',
                        'commands': {
                            'class': {
                                'meta': 'A generic hook manager for Classes',
                                'exec': ios_generate.clazz
                            },
                            'simple': {
                                'meta': 'Simple hooks for each Class method',
                                'exec': ios_generate.simple
                            }
                        },
                    }
                }
            },
            'pasteboard': {
                'meta': 'Work with the iOS pasteboard',
                'commands': {
                    'monitor': {
                        'meta': 'Monitor the iOS pasteboard',
                        'exec': pasteboard.monitor
                    }
                }
            },
            'sslpinning': {
                'meta': 'Work with iOS SSL pinning',
                'commands': {
                    'disable': {
                        'meta': 'Attempt to disable SSL pinning in various iOS libraries/classes',
                        'flags': ['--quiet'],
                        'exec': ios_pinning.ios_disable
                    }
                }
            },
            'jailbreak': {
                'meta': 'Work with iOS Jailbreak detection',
                'commands': {
                    'disable': {
                        'meta': 'Attempt to disable Jailbreak detection',
                        'exec': jailbreak.disable
                    },
                    'simulate': {
                        'meta': 'Attempt to simulate a Jailbroken environment',
                        'exec': jailbreak.simulate
                    },
                }
            },
            'monitor': {
                'meta': 'Commands to work with ios function monitoring',
                'commands': {
                    'crypto': {
                        'meta': 'Monitor CommonCrypto operations',
                        'exec': ios_crypto.crypto_enable
                    }
                },
            },
        }
    },

    'exit': {
        'meta': 'Exit',
    },
}



import json
import os
from typing import List

import click
from tabulate import tabulate

from objection.state.connection import state_connection
from ..utils.helpers import clean_argument_flags
from ..utils.helpers import sizeof_fmt, pretty_concat

BLOCK_SIZE = 40960000


def _is_string_input(args: list) -> bool:
    """
        Checks if --string is in the list of tokens received form the
        command line.

        :param args:
        :return:
    """

    return len(args) > 0 and '--string' in args


def _should_only_dump_offsets(args: list) -> bool:
    """
        Checks if --offsets-only is in the list pf tokens received
        from the command line.

        :param args:
        :return:
    """

    return '--offsets-only' in args


def _is_string_pattern(args: list) -> bool:
    """
        Checks if --string-pattern is in the list of tokens received form the
        command line.

        :param args:
        :return:
    """

    return len(args) > 0 and '--string-pattern' in args


def _is_string_replace(args: list) -> bool:
    """
        Checks if --string-replace is in the list of tokens received form the
        command line.

        :param args:
        :return:
    """

    return len(args) > 0 and '--string-replace' in args


def _should_output_json(args: list) -> bool:
    """
        Checks if --json is in the list of tokens received from the command line.

        :param args:
        :return:
    """

    return len(args) > 0 and '--json' in args


def _get_chunks(addr: int, size: int, block_size: int = BLOCK_SIZE) -> List:
    """
        Determine chunks of

        :param addr:
        :param size:
        :param block_size:
        :return:
    """

    if size < block_size:
        return [(addr, size)]

    block_count = size // block_size
    extra_block = size % block_size
    current_address = addr
    ranges = []

    for i in range(block_count):
        ranges.append((current_address, block_size))
        current_address += block_size

    if extra_block != 0:
        ranges.append((current_address, extra_block))

    return ranges


# TODO: Dump memory on hooked methods.
# A PR in the repo this method is based on has an idea for this
#
# https://github.com/Nightbringer21/fridump/pull/3

def dump_all(args: list) -> None:
    """
        Dump memory from the currently injected process.
        Loosely based on:
            https://github.com/Nightbringer21/fridump

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: memory dump all <local destination>', bold=True)
        return

    # the destination file to write the dump to
    destination = args[0]

    # Check for file override
    if os.path.exists(destination):
        click.secho('Destination file {dest} already exists'.format(dest=destination), fg='yellow', bold=True)
        if not click.confirm('Continue, appending to the file?'):
            return

    # access type used when enumerating ranges
    access = 'rw-'

    api = state_connection.get_api()
    ranges = api.memory_list_ranges(access)

    total_size = sum([x['size'] for x in ranges])
    click.secho('Will dump {0} {1} images, totalling {2}'.format(
        len(ranges), access, sizeof_fmt(total_size)), fg='green', dim=True)

    with click.progressbar(ranges) as bar:
        for image in bar:
            dump = bytearray()
            bar.label = 'Dumping {0} from base: {1}'.format(sizeof_fmt(image['size']), hex(int(image['base'], 16)))

            # catch and exception thrown while dumping.
            # this could for a few reasons like if the protection
            # changes or the range is reallocated
            try:
                # grab the (size) bytes starting at the (base_address) in chunks of BLOCK_SIZE
                chunks = _get_chunks(int(image['base'], 16), int(image['size']), BLOCK_SIZE)
                for chunk in chunks:
                    dump.extend(bytearray(api.memory_dump(chunk[0], chunk[1])))

            except Exception as e:
                continue

            # append the results to the destination file
            with open(destination, 'ab') as f:
                f.write(dump)

    click.secho('Memory dumped to file: {0}'.format(destination), fg='green')


def dump_from_base(args: list) -> None:
    """
        Dump memory from a base address for a specific size to file

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) < 3:
        click.secho('Usage: memory dump from_base <base_address> <size_to_dump> <local_destination>', bold=True)
        return

    # the destination file to write the dump to
    base_address = args[0]
    memory_size = args[1]
    destination = args[2]

    # Check for file override
    if os.path.exists(destination):
        click.secho('Destination file {dest} already exists'.format(dest=destination), fg='yellow', bold=True)
        if not click.confirm('Override?'):
            return

    click.secho('Dumping {0} from {1} to {2}'.format(sizeof_fmt(int(memory_size)), base_address, destination),
                fg='green', dim=True)

    api = state_connection.get_api()

    # iirc, if you don't cast the return type to a bytearray it uses the sizeof(int) per cell, which is massive
    dump = bytearray()
    chunks = _get_chunks(int(base_address, 16), int(memory_size), BLOCK_SIZE)
    for chunk in chunks:
        dump.extend(bytearray(api.memory_dump(chunk[0], chunk[1])))

    # append the results to the destination file
    with open(destination, 'wb') as f:
        f.write(dump)

    click.secho('Memory dumped to file: {0}'.format(destination), fg='green')


def list_modules(args: list = None) -> None:
    """
        List modules loaded in the current process.

        :param args:
        :return:
    """

    if _should_output_json(args) and len(args) < 2:
        click.secho('Usage: memory list modules (--json <local destination>)', bold=True)
        return

    if not _should_output_json(args):
        click.secho('Save the output by adding `--json modules.json` to this command', dim=True)

    api = state_connection.get_api()
    modules = api.memory_list_modules()

    if _should_output_json(args):
        destination = args[args.index('--json') + 1]

        click.secho('Writing modules as json to {0}...'.format(destination), dim=True)

        with open(destination, 'w') as f:
            f.write(json.dumps(modules, indent=2))

        click.secho('Wrote modules to: {0}'.format(destination), fg='green')
        return

    # Just dump it to the screen
    click.secho(tabulate(
        [[
            entry['name'],
            entry['base'],
            str(entry['size']) + ' (' + sizeof_fmt(entry['size']) + ')',
            pretty_concat(entry['path']),
        ] for entry in modules], headers=['Name', 'Base', 'Size', 'Path'],
    ))


def list_exports(args: list) -> None:
    """
        Dumps the exported methods from a loaded module to screen.

        :param args:
        :return:
    """

    if _should_output_json(args) and len(args) < 3:
        click.secho('Usage: memory list exports <module name> (--json <local destination>)', bold=True)
        return

    if not _should_output_json(args):
        click.secho('Save the output by adding `--json exports.json` to this command', dim=True)

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: memory list exports <module name>', bold=True)
        return

    module_to_list = args[0]

    api = state_connection.get_api()
    exports = api.memory_list_exports(module_to_list)

    if _should_output_json(args):
        destination = args[args.index('--json') + 1]

        click.secho('Writing exports as json to {0}...'.format(destination), dim=True)

        with open(destination, 'w') as f:
            f.write(json.dumps(exports, indent=2))

        click.secho('Wrote exports to: {0}'.format(destination), fg='green')
        return

    # Just dump it to the screen
    click.secho(tabulate(
        [[
            entry['type'],
            entry['name'],
            entry['address'],
        ] for entry in exports], headers=['Type', 'Name', 'Address'],
    ))


def find_pattern(args: list) -> None:
    """
        Searches the current processes accessible memory for a specific pattern.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: memory search "<pattern eg: 41 41 41 ?? 41>" (--string) (--offsets-only)', bold=True)
        return

    # if we got a string as input, convert it to hex
    if _is_string_input(args):
        pattern = ' '.join(hex(ord(x))[2:] for x in args[0])
    else:
        pattern = args[0]

    click.secho('Searching for: {0}'.format(pattern), dim=True)

    api = state_connection.get_api()
    data = api.memory_search(pattern, _should_only_dump_offsets(args))

    if len(data) > 0:
        click.secho('Pattern matched at {0} addresses'.format(len(data)), fg='green')
        if _should_only_dump_offsets(args):
            for address in data:
                click.secho(address)

    else:
        click.secho('Unable to find the pattern in any memory region')


def replace_pattern(args: list) -> None:
    """
        Searches the current processes accessible memory for a specific pattern and replaces it with given bytes or string.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) < 2:
        click.secho('Usage: memory replace "<search pattern eg: 41 41 ?? 41>" "<replace value eg: 41 50>" (--string-pattern) (--string-replace)', bold=True)
        return

    # if we got a string as search pattern input, convert it to hex
    if _is_string_pattern(args):
        pattern = ' '.join(hex(ord(x))[2:] for x in args[0])
    else:
        pattern = args[0]

    # if we got a string as replace input, convert it to int[], otherwise convert hex to int[]
    replace = args[1]
    if _is_string_replace(args):
        replace = [ord(x) for x in replace]
    else:
        replace = [int(x, 16) for x in replace.split(' ')]

    click.secho('Searching for: {0}, replacing with: {1}'.format(pattern, args[1]), dim=True)

    api = state_connection.get_api()
    data = api.memory_replace(pattern, replace)

    if len(data) > 0:
        click.secho('Pattern replaced at {0} addresses'.format(len(data)), fg='green')
        for address in data:
            click.secho(address)

    else:
        click.secho('Unable to find the pattern in any memory region')


def write(args: list) -> None:
    """
        Write an arbitrary amount of bytes to an arbitrary memory address.

        Needless to say, use with caution. =P

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) < 2:
        click.secho('Usage: memory write "<address>" "<pattern eg: 41 41 41 41>" (--string)', bold=True)
        return

    destination = args[0]
    pattern = args[1]

    if _is_string_input(args):
        pattern = [ord(x) for x in pattern]
    else:
        pattern = [int(x, 16) for x in pattern.split(' ')]

    click.secho('Writing byte array: {0} to {1}'.format(pattern, destination), dim=True)

    api = state_connection.get_api()
    api.memory_write(destination, pattern)



class Device(object):
    """ Represents a mobile device """
    pass


class Android(Device):
    """ Represents Android specific configurations. """

    name = 'android'
    path_separator = '/'


class Ios(Device):
    """ Represents iOS specific configurations. """

    name = 'ios'
    path_separator = '/'


class DeviceState(object):
    """ A class representing the state of a device and its runtime. """

    platform: Device
    version: str

    def set_version(self, v: str):
        """
            Set the running OS version

            :param v:
            :return:
        """

        self.version = v

    def set_platform(self, t: Device):
        """
            Set's the device type

            :param t:
            :return:
        """

        self.platform = t

    def __repr__(self) -> str:
        return f'<Type: {self.platform} >'


device_state = DeviceState()



import sys
from importlib import metadata
from pathlib import Path

import tomllib


def _load_version() -> str:
    """
        Prefer the installed package metadata and fall back to pyproject.toml
        when running from a checkout.
    """

    try:
        return metadata.version("objection")
    except metadata.PackageNotFoundError:
        pyproject_path = Path(__file__).resolve().parent.parent / "pyproject.toml"
        try:
            with pyproject_path.open("rb") as f:
                return tomllib.load(f)["project"]["version"]
        except Exception:
            return "0.0.0"


__version__ = _load_version()

# helper containing a python 3 related warning
# if this is run with python 2
if sys.version_info < (3,):
    raise ImportError(
        '''
    You are running objection {0} on Python 2

    Unfortunately objection {0} and above are not compatible with Python 2.
    That's a bummer; sorry about that.  Make sure you have Python 3, pip >= and
    setuptools >= 24.2 to avoid these kinds of issues in the future:

     $ pip install pip setuptools --upgrade

    You could also setup a virtual Python 3 environment.

     $ pip install pip setuptools --upgrade
     $ pip install virtualenv
     $ virtualenv --python=python3 ~/virt-python3
     $ source ~/virt-python3/bin/activate

    This will make an isolated Python 3 installation available and active, ready
    to install and use objection.
    '''.format(__version__))



import pprint

import click
from prompt_toolkit import prompt
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.javascript import JavascriptLexer
from tabulate import tabulate

from objection.state.connection import state_connection


def _should_ignore_methods_with_arguments(args) -> bool:
    """
        Check if the --without-arguments flag exists

        :param args:
        :return:
    """

    return len(args) > 0 and '--without-arguments' in args


def _should_print_as_utf8(args) -> bool:
    """
        Check if the --to-utf8 flag exists

        :param args:
        :return:
    """

    return len(args) > 0 and '--to-utf8' in args


def _should_return_as_string(args) -> bool:
    """
        Check if the --return-string flag exists

        :param args:
        :return:
    """

    return len(args) > 0 and '--return-string' in args


def _should_interpret_inline_js(args) -> bool:
    """
        Check if we have the --inline flag

        :param args:
        :return:
    """

    return len(args) > 0 and '--inline' in args


def instances(args: list) -> None:
    """
        Asks the agent to print the currently live instances of a particular class

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: ios heap search instances <class> (eg: com.example.test)', bold=True)
        return

    target_class = args[0]

    api = state_connection.get_api()
    instance_results = api.ios_heap_print_live_instances(target_class)

    # export interface IHeapObject {
    #   className: string;
    #   handle: string;
    #   ivars: any[string];
    #   kind: string;
    #   methods: string[];
    #   superClass: string;
    # }

    if len(instance_results) <= 0:
        return

    click.secho(tabulate(
        [[
            entry['handle'],
            entry['kind'],
            entry['className'],
            entry['superClass'],
            len(entry['ivars']),
            len(entry['methods'])
        ] for entry in instance_results], headers=['Handle', 'Kind', 'Class', 'Super', 'iVars', 'Methods'],
    ))


def ivars(args: list) -> None:
    """
        Get ivars for an Objective-C object at a pointer

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: ios heap print ivars <pointer> (eg: 0x600001130660)', bold=True)
        return

    target_pointer = args[0]

    api = state_connection.get_api()
    ivar_results = api.ios_heap_print_ivars(target_pointer, _should_print_as_utf8(args))

    click.secho(tabulate(
        [[
            key, value
        ] for key, value in ivar_results[1].items()], headers=['iVar', 'Value'],
    ))


def methods(args: list) -> None:
    """
        Get methods for an Objective-C object at a pointer

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: ios heap print methods <pointer> (eg: 0x600001130660)', bold=True)
        return

    target_pointer = args[0]

    api = state_connection.get_api()
    method_results = api.ios_heap_print_methods(target_pointer)

    # apply argument filters
    if _should_ignore_methods_with_arguments(args):
        method_results[1] = list(filter(lambda x: ':' not in x, method_results[1]))

    click.secho(tabulate(
        [[
            entry,
            entry.split(" ")[0],
            "{type} [{clazz} {method}]".format(  # hacky, right? :D
                type=entry.split(" ")[0], clazz=method_results[0], method=entry.split(" ")[1])
        ] for entry in method_results[1]], headers=['Method', 'Type', 'Full'],
    ))


def execute(args: list) -> None:
    """
        Executes a method on a pointer which is assumed to be an Objective-C
        object.

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: ios heap execute method <pointer> <method> (eg: 0x600001130660)', bold=True)
        return

    target_pointer = args[0]
    method = args[1]

    if ':' in method:
        click.secho('Unfortunately, only methods that do not require arguments are supported.', fg='yellow')
        return

    api = state_connection.get_api()
    exec_results = api.ios_heap_exec_method(target_pointer, method, _should_return_as_string(args))

    click.secho(pprint.pformat(exec_results))


def evaluate(args: list) -> None:
    """
        Evaluate JavaScript on an Objective-C pointer.

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: ios heap execute js <pointer> (eg: 0x600001130660) ' +
                    '(optional: --inline) (optional: <JavaScript source>)', bold=True)
        return

    target_pointer = args[0]

    # adding the --inline flag would trigger reading the line contents
    # as JavaScript sources
    if _should_interpret_inline_js(args):
        args.remove('--inline')
        js = ''.join(args[1:])

        click.secho('Reading inline JavaScript for evaluation...', dim=True)
        click.secho('{}\n'.format(js), fg='green', dim=True)

    else:
        js = prompt(
            click.secho('(The pointer at `{pointer}` will be available as the `ptr` variable.)n'.format(
                pointer=target_pointer
            ), dim=True),
            multiline=True, lexer=PygmentsLexer(JavascriptLexer),
            bottom_toolbar='JavaScript edit mode. [ESC] and then [ENTER] to accept. [CTRL] + C to cancel.').strip()

        click.secho('JavaScript capture complete. Evaluating...', dim=True)

    api = state_connection.get_api()
    api.ios_heap_evaluate_js(target_pointer, js)



class AppState(object):
    """ A class representing generic state variable for this app """

    def __init__(self):
        self.debug_hooks = False
        self.debug = False
        self.api_host = '127.0.0.1'
        self.api_port = 8888
        self.successful_commands = []

    def add_command_to_history(self, command: str) -> None:
        """
            Adds a command to the list of successful commands.

            :param command:
            :return:
        """

        if command not in self.successful_commands:
            self.successful_commands.append(command)

    def clear_command_history(self) -> None:
        """
            Clears the list of successful commands recorded
            for this session.

            :return:
        """

        self.successful_commands = []

    def should_debug_hooks(self) -> bool:
        """
            Returns if debugging of Frida hooks is needed.

            :return:
        """

        return self.debug_hooks

    def should_debug(self) -> bool:
        """

            Checks if debugging is enabled

            :return:
        """

        return self.debug


app_state = AppState()



import os

import click

from objection.state.connection import state_connection


def clazz(args: list) -> None:
    """
        Simply echoes the source for a generic Hook Manager
        sample for Objective-C hooks with Frida.

        :param args:
        :return:
    """

    js_path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        '../../utils/assets', 'objchookmanager.js'
    )

    with open(js_path, 'r') as f:
        click.secho(f.read(), dim=True)


def simple(args: list) -> None:
    """
        Generate simple hooks for all methods in a class.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: ios hooking generate simple <class name>', bold=True)
        return

    classname = args[0]

    api = state_connection.get_api()
    methods = api.ios_hooking_get_class_methods(classname, False)

    if len(methods) <= 0:
        click.secho('No class / methods found')
        return

    click.secho("var target = ObjC.classes.{};".format(classname), dim=True)

    for method in methods:
        hook = """
Interceptor.attach(target['{method}'].implementation, {
  onEnter: function (args) {
    console.log('Entering {method}!');
  },
  onLeave: function (retval) {
    console.log('Leaving {method}');
  },
});
    """.replace('{method}', method)

        click.secho(hook, dim=True)



import click
from tabulate import tabulate

from objection.state.connection import state_connection
from ..state.jobs import job_manager_state, Job


def show(args: list = None) -> None:
    """
        Show all of the jobs that are currently running

        :return:
    """

    sync_job_manager()
    jobs = job_manager_state.jobs

    # click.secho(tabulate(
    #     [[
    #         entry['uuid'],
    #         sum([
    #             len(entry[x]) for x in [
    #                 'invocations', 'replacements', 'implementations'
    #             ] if x in entry
    #         ]),
    #         entry['type'],
    #     ] for entry in jobs], headers=['Job ID', 'Hooks', 'Name'],
    # ))
    click.secho(tabulate(
        [[
            uuid,
            job.job_type,
            job.name,
        ] for uuid, job in jobs.items()], headers=['Job ID', 'Type', 'Name'],
    ))


def kill(args: list) -> None:
    """
        Kills a specific objection job.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: jobs kill <uuid>', bold=True)
        return

    job_uuid = int(args[0])

    job_manager_state.remove_job(job_uuid)


def list_current_jobs() -> dict:
    """
        Return a list of the currently listed objection jobs.
        Used for tab completion in the repl.
    """

    sync_job_manager()
    resp = {}

    for uuid, job in job_manager_state.jobs.items():
        resp[str(uuid)] = str(uuid)

    return resp


def sync_job_manager() -> dict[int, Job]:
    try:
        api = state_connection.get_api()
        jobs = api.jobs_get()

        for job in jobs:
            job_uuid = int(job['identifier'])
            job_name = job['type']
            if job_uuid not in job_manager_state.jobs:
                job_manager_state.jobs[job_uuid] = Job(job_name, 'hook', None, job_uuid)

        return job_manager_state.jobs
    except:
        print("REPL not ready")




import click

from objection.state.connection import state_connection
from objection.utils.helpers import clean_argument_flags

def _should_dump_backtrace(args: list = None) -> bool:
    """
        Check if --dump-backtrace is part of the arguments.

        :param args:
        :return:
    """

    return '--dump-backtrace' in args

def analyze_implicit_intents(args: list) -> None:
    """
        Analyzes implicit intents in hooked methods.
    """
    api = state_connection.get_api()
    should_backtrace = _should_dump_backtrace(args)

    api.android_intent_analyze(should_backtrace)
    if not should_backtrace:
        click.secho('Started implicit intent analysis', bold=True)
    else:
        click.secho('Started implicit intent analysis with backtrace', bold=True)


def launch_activity(args: list) -> None:
    """
        Launches an activity class using an Android Intent

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) < 1:
        click.secho('Usage: android intent launch_activity <activity_class>', bold=True)
        return

    intent_class = args[0]

    api = state_connection.get_api()
    api.android_intent_start_activity(intent_class)


def launch_service(args: list) -> None:
    """
        Launches an exported service using an Android Intent

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) < 1:
        click.secho('Usage: android intent launch_service <service_class>', bold=True)
        return

    intent_class = args[0]

    api = state_connection.get_api()
    api.android_intent_start_service(intent_class)
